"""
A module that provides high-level functions to perform actions on UAVs.
"""
import logging
import time
from copy import deepcopy
from typing import Optional

from pymavlink.dialects.v20.ardupilotmega import (
    MAV_CMD_COMPONENT_ARM_DISARM,
    MAV_CMD_DO_REPOSITION,
    MAV_CMD_DO_SET_SERVO,
    MAV_MODE_FLAG_SAFETY_ARMED,
)

from albatros.enums import ConnectionType
from albatros.telem import ComponentAddress
from albatros.telem.drivers import (
    DirectConnectionDriver,
    RedisConnectionDriver,
    TelemDriver,
    TestDriver,
)

from .enums import CommandResult
from .nav.position import PositionGPS, PositionNED, ned2geo
from .outgoing.commands import get_command_int_message, get_command_long_message
from .outgoing.param_messages import (
    get_param_request_list_message,
    get_param_request_read_message,
    get_param_set_message,
)
from .telem.message_models import CommandACK, Heartbeat, ParamValue
from .telem.uav_data import UAVData

logger = logging.getLogger(__name__)


TIMEOUT_S = 0.25
HEARTBEAT_TIMEOUT_S = 1.25
CHECK_PERIOD_S = 0.01


class UAV:
    """Class that provides actions that the UAV can perform.
    Actions are common for aircraft and copter vehicle type.
    """

    def __init__(
        self,
        uav_addr: ComponentAddress = ComponentAddress(system_id=1, component_id=1),
        my_addr: ComponentAddress = ComponentAddress(system_id=1, component_id=191),
        connection_type: ConnectionType = ConnectionType.DIRECT,
        device: Optional[str] = "udpin:0.0.0.0:14550",
        baud_rate: Optional[int] = 115200,
        host: Optional[str] = None,
        port: Optional[int] = None,
    ) -> None:
        self.uav_addr = uav_addr
        self.my_addr = my_addr
        self.driver: TelemDriver
        self.data = UAVData(uav_addr)

        if connection_type == ConnectionType.DIRECT:
            if not device or not baud_rate:
                raise ValueError
            self.driver = DirectConnectionDriver(my_addr, self.data, device, baud_rate)

        elif connection_type == ConnectionType.REDIS:
            if not host or not port:
                raise ValueError
            self.driver = RedisConnectionDriver(my_addr, self.data, host, port)

        elif connection_type == ConnectionType.TEST:
            self.driver = TestDriver(my_addr, self.data)
        else:
            raise NotImplementedError

        if not self.driver.make_connection():
            raise TimeoutError

    def is_armed(self) -> bool:
        """Check whether the UAV is armed."""
        armed_flag = self.data.heartbeat.base_mode & MAV_MODE_FLAG_SAFETY_ARMED
        return bool(armed_flag)

    def wait_gps_fix(self) -> None:
        """Wait for GPS 3D fix."""
        while self.data.gps_raw_int.fix_type < 3 or self.data.gps_raw_int.lat == 0:
            time.sleep(CHECK_PERIOD_S)

    def wait_heartbeat(self) -> Heartbeat:
        """Wait for next heartbeat message."""
        clock_start = time.time()
        while True:
            time_dif = time.time() - self.data.heartbeat.timestamp_ms / 1_000.0
            if time_dif < CHECK_PERIOD_S:
                heartbeat = deepcopy(self.data.heartbeat)
                self.data.heartbeat.timestamp_ms = 0
                return heartbeat
            time.sleep(CHECK_PERIOD_S)
            if time.time() - clock_start > HEARTBEAT_TIMEOUT_S:
                raise TimeoutError

    def wait_command_ack(self) -> CommandACK:
        """Wait for command execution status.

        :returns: CommandACK object containing a execution result.
        """
        clock_start = time.time()
        while True:
            time_dif = time.time() - self.data.command_ack.timestamp_ms / 1_000.0
            if time_dif < CHECK_PERIOD_S:
                command_ack = deepcopy(self.data.command_ack)
                self.data.command_ack.timestamp_ms = 0
                return command_ack
            time.sleep(CHECK_PERIOD_S)
            if time.time() - clock_start > TIMEOUT_S:
                raise TimeoutError

    def wait_param(self) -> ParamValue:
        """Wait for requested param value.

        :returns: ParamValue object.
        """
        clock_start = time.time()
        while True:
            time_dif = time.time() - self.data.param_value.timestamp_ms / 1_000.0
            if time_dif < CHECK_PERIOD_S:
                param = deepcopy(self.data.param_value)
                self.data.param_value.timestamp_ms = 0
                return param
            time.sleep(CHECK_PERIOD_S)
            if time.time() - clock_start > TIMEOUT_S:
                raise TimeoutError

    def arm(self) -> bool:
        """Arms motors."""
        msg = get_command_long_message(
            target_system=self.uav_addr.system_id,
            target_component=self.uav_addr.component_id,
            command=MAV_CMD_COMPONENT_ARM_DISARM,
            param1=1,
        )

        self.driver.send(msg)
        logger.info("Arm command sent.")
        self.wait_heartbeat()

        return self.is_armed()

    def disarm(self) -> bool:
        """Disarms motors."""
        msg = get_command_long_message(
            target_system=self.uav_addr.system_id,
            target_component=self.uav_addr.component_id,
            command=MAV_CMD_COMPONENT_ARM_DISARM,
            param1=0,
        )

        self.driver.send(msg)
        logger.info("Disarm command sent.")

        if self.wait_command_ack().result != CommandResult.ACCEPTED:
            return False
        return True

    def set_servo(self, instance_number: int, pwm: int) -> bool:
        """Set a servo to a desired PWM value.

        :param instance_number: servo number.
        :param pwm: PWM value to set.
        """
        msg = get_command_long_message(
            target_system=self.uav_addr.system_id,
            target_component=self.uav_addr.component_id,
            command=MAV_CMD_DO_SET_SERVO,
            param1=instance_number,
            param2=pwm,
        )

        self.driver.send(msg)
        logger.info("Set servo command sent.")

        if self.wait_command_ack().result != CommandResult.ACCEPTED:
            return False
        return True

    def fly_to_gps_position(self, lat_int: int, lon_int: int, alt_m: float) -> bool:
        """Reposition the vehicle to a specific WGS84 global position.

        :param lat_int: Latitude of the target point.
        :param lon_int: Longitude of the target point.
        :param alt_m: Altitude of the target point in meters.

        Works only in Guided mode.
        """
        msg = get_command_int_message(
            target_system=self.uav_addr.system_id,
            target_component=self.uav_addr.component_id,
            command=MAV_CMD_DO_REPOSITION,
            x=lat_int,
            y=lon_int,
            z=alt_m,
        )

        self.driver.send(msg)
        logger.info("Flight to point command sent.")

        if self.wait_command_ack().result != CommandResult.ACCEPTED:
            return False
        return True

    def get_corrected_position(self) -> PositionGPS:
        """
        :returns: the vehicle position corrected for the distance the vehicle traveled after the message was received.
        """
        movement_time = time.time() - self.data.global_position_int.timestamp_ms / 1000
        north_shift = movement_time * self.data.global_position_int.vx / 100
        east_shift = movement_time * self.data.global_position_int.vy / 100
        z_shift = movement_time * self.data.global_position_int.vz / 100
        corrected_altitude = z_shift + self.data.global_position_int.relative_alt / 1000
        last_known_position = PositionGPS(
            self.data.global_position_int.lat,
            self.data.global_position_int.lon,
        )
        shift_vector = PositionNED(north_shift, east_shift, corrected_altitude)
        return ned2geo(last_known_position, shift_vector)

    def fetch_one_param(self, param_id: str) -> ParamValue:
        """Fetch single parameter from UAV"""
        msg = get_param_request_read_message(
            target_system=self.uav_addr.system_id,
            target_component=self.uav_addr.component_id,
            param_id=param_id.encode("ascii"),
            param_index=-1,
        )

        self.driver.send(msg)
        logger.debug("Param request read message sent.")
        return self.wait_param()

    def request_all_parameters(self) -> None:
        """
        Send a command to request values of every parameter from the uav.
        If you need specific parameters, you should use request_one_parameter instead
        """
        msg = get_param_request_list_message(
            target_system=self.uav_addr.system_id,
            target_component=self.uav_addr.component_id,
        )

        self.driver.send(msg)
        logger.debug("Param request list message sent.")

    def set_parameter(self, param_id: str, new_value: float) -> bool:
        """
        Set a parameter to the specified value
        """
        msg = get_param_set_message(
            target_system=self.uav_addr.system_id,
            target_component=self.uav_addr.component_id,
            param_id=param_id.encode("ascii"),
            param_value=new_value,
        )

        self.driver.send(msg)
        logger.debug("Param set message sent.")
        return abs(self.wait_param().param_value - new_value) < 0.0001
